import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.net.*;
import java.awt.*;
import java.util.*;
import java.io.*;

/**
   TODO: ensure that simulators die when complete

   Graphical display and management of Darwin Game tournaments.

   @sa Darwin for single matches and debugging.

   <pre>
     java Tournament <i>mapfile</i> Creature0 Creature1 ...
   </pre>

   If mapfile begins with "mz_" it is assumed to have one creature on it.  If
   mapfile begins with "ns_" it is assumed to have two creatures on it.

   <p> The Tournament class runs each trial with a fresh instance of
   the classes, intentionally reloading them from disk.  This prevents
   classes from sharing any information with themselves between runs
   using static variables.
 */
public class Tournament extends JFrame implements Icon {

    /** Description of a creature for the scoreboard. */
    private abstract class Description implements Comparable<Description> {
        public String className;
        public String authorName;
        public Icon   icon;

        public ArrayList<Integer> times = new ArrayList<Integer>();
        
        public Description(String cname) throws ClassNotFoundException, 
                                                InstantiationException, 
                                                IllegalAccessException,
                                                IOException {
            className = cname;
            Class c = Simulator.loadClass(className);
            assert c != null;
            Creature instance = (Creature)c.newInstance();
            authorName = instance.getAuthorName();

            icon = new ImageIcon(Simulator.getImage(c, Direction.SOUTH));
        }

        abstract public int compareTo(Description obj);

        protected Simulator.Result runTrial(String mapName, 
                                            Class[] creatures) {
            Simulator simulator = new Simulator(mapName, creatures);
            simulator.setView3D(false);

            // Launch the simulation
            if (isMaze) {
                simulator.setDelayTime(10000);
            } else {
                // Need a long enough delay time to
                // avoid punishing creatures that use
                // synchronization heavily.  
                simulator.setDelayTime(8 * 100000);
            }

            synchronized (Tournament.this) {
                currentSimulator = simulator;
            }
            Tournament.this.repaint();

            // Run until completion.
            Simulator.Result result = simulator.getResult();
            int count = 0;
            final int sleeptime = 25; // ms
            while (result == null) {
                try {
                    // Don't sleep too long; if we don't notice
                    // that the game has ended we might let it run
                    // for too long past the end, which affects 
                    // maze running timings.
                    Thread.sleep(sleeptime);
                    ++count;
                    if (count > 1500 / sleeptime) {
                        // Repaint once per 1.5 seconds
                        Tournament.this.repaint();
                        count = 0;
                    }
                } catch (InterruptedException e) {}
                result = simulator.getResult();
            }

            // Force the icon to repaint, showing the result
            Tournament.this.repaint();
            simulator.stop();
            return result;
        }
    }
    
    /** For maze tournaments */
    private class MZDescription extends Description {
        public int    bestTime = Integer.MAX_VALUE;
        public int    worstTime = 0;
        public int    meanTime = Integer.MAX_VALUE;

        public ArrayList<Integer> times = new ArrayList<Integer>();
        
        public MZDescription(String cname) throws ClassNotFoundException, 
                                                InstantiationException, 
                                                IllegalAccessException,
                                                IOException {
            super(cname);
        }

        public int compareTo(Description obj) {
            MZDescription that = (MZDescription)obj;
            return this.meanTime - that.meanTime;
        }
        
        /** Run one trial and update this description with the result. */
        public void runTrial(String mapName) {
            try {
                showText(className + ": ");
                Class[] creatures = {Simulator.loadClass(className)};
                Simulator.Result result = runTrial(mapName, creatures);
                
                addTime(result.timeSteps);

                if (result.species != creatures[0]) {
                    showText("" + result.timeSteps + " (timeout)\n");
                } else {
                    showText("" + result.timeSteps + "\n");
                }


                
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        /** Update times to reflect this change. */
        public void addTime(int time) {
            bestTime = Math.min(bestTime, time);
            worstTime = Math.max(worstTime, time);
            times.add(time);
            meanTime = 0;
            for (int t : times) {
                meanTime += t;
            }
            meanTime /= times.size();
        }
    }

    final static private Random rand = new Random();

    /** For natural selection tournaments */
    private class NSDescription extends Description {
        public int    bestTime = Integer.MAX_VALUE;
        public int    worstTime = 0;
        public int    meanTime = Integer.MAX_VALUE;

        public int    win;
        public int    lose;
        public int    draw;

        public int    score;
        
        public NSDescription(String cname) throws ClassNotFoundException, 
                                                InstantiationException, 
                                                IllegalAccessException,
                                                IOException {
            super(cname);
        }

        public int compareTo(Description obj) {
            NSDescription that = (NSDescription)obj;
            int c = that.score - this.score;
            if (c == 0) {
                // Tie; see who has more wins
                return that.win - this.win;
            } else {
                return c;
            }
        }
        
        /** Run one trial and update this description (and the opponent's) with the result. */
        public void runTrial(String mapName, NSDescription opponent) {
            try {
                Class[] creatures = {Simulator.loadClass(className), 
                                     Simulator.loadClass(opponent.className)};
                
                showText(className + " vs. " + opponent.className + ": ");

                // For fairness, change up who is first
                if (rand.nextInt(2) == 0) {
                    // Swap
                    Class temp = creatures[0];
                    creatures[0] = creatures[1];
                    creatures[1] = temp;
                }

                Simulator.Result result = runTrial(mapName, creatures);

                String outcome = className + " vs. " + opponent.className + ": " + 
                    result.result + " " + result.why;

                System.out.println(outcome);
                showText(result.result + " " + result.why + "\n");

                processResult(result);
                opponent.processResult(result);
                
            } catch (ClassNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }
        }

        private void processResult(Simulator.Result result) {
            switch (result.code) {
            case ALL_EXTINCT:
                // No score change
                ++draw;
                break;
                
            case TIME_ELAPSED:
                if (result.species == null) {
                    // No majority
                    ++draw;
                    score += 1;
                } else if (result.species.getName().equals(className)) {
                    // I won by majority
                    ++win;
                    score += 2;
                } else {
                    // I lost by majority
                    ++lose;
                    score += 1;
                }
                break;
                
            case DOMINATION:
                if (result.species.getName().equals(className)) {
                    // I won by domination
                    ++win;
                    score += 3;
                } else {
                    ++lose;
                    // No score change
                }
                break;
            }
        }
    }

    private boolean isMaze;
    private Description[] creatureArray;
    private Description[] sortedCreatureArray;
    private String mapFilename;
    private ResultData resultData;
    private JProgressBar progressBar;
    private int currentTrial;
    private int numTrials;

    private Simulator currentSimulator;
    private JTextArea resultArea;
    private JScrollPane resultScroll;
    private JLabel resultPicture;

    private static final int NUM_MZ_REPEATS = 8;
    private static final int NUM_NS_REPEATS = 4;

    private Tournament(String mapfile, String[] creatureClassNames) {
        super("Darwin Game Tournament - " + mapfile);
 
        System.setSecurityManager(new MaximumSecurityManager(new String[]{"readFileDescriptor"}));

        isMaze = mapfile.contains("mz_");

        creatureArray = new Description[creatureClassNames.length];
        for (int i = 0; i < creatureArray.length; ++i) {
            try {
                
                if (isMaze) {
                    creatureArray[i] = new MZDescription(creatureClassNames[i]);
                } else {
                    creatureArray[i] = new NSDescription(creatureClassNames[i]);
                }

            } catch (Exception e) {

                System.err.println("Error while loading " + creatureClassNames[i] + ":\n" + e);
                e.printStackTrace();
                System.exit(-1);

            }
        }
        sortedCreatureArray = (Description[])creatureArray.clone();
        mapFilename = mapfile;

        // Must happen before GUI is created.
        currentTrial = 0;
        if (isMaze) {
            numTrials = NUM_MZ_REPEATS * creatureArray.length;
        } else {
            // Try every creature against every other
            numTrials = NUM_NS_REPEATS * (creatureArray.length * (creatureArray.length - 1)) / 2;

            // Update table column names
            columnNames[WLD_COL] = "W:D:L";
            columnNames[SCORE_COL] = "Score";
        }

        makeGUI();
        pack();
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setVisible(true);

        if (isMaze) {
            runAllMZTrials(mapfile);
        } else {
            runAllNSTrials(mapfile);
        }

        progressBar.setVisible(false);
        System.out.println("Trials complete");
    }

    public int getIconHeight() {
        return 500;
    }

    public int getIconWidth() {
        return 550;
    }

    public synchronized void paintIcon(Component c, Graphics _g, int tx, int ty) {
        if (currentSimulator != null) {
            currentSimulator.paintIcon(c, _g, tx, ty);
        }
    }

    private void runAllNSTrials(String mapfile) {
        for (int k = 0; k < NUM_NS_REPEATS; ++k) {
            for (int i = 0; i < creatureArray.length; ++i) {
                for (int j = 0; j < i; ++j) {
                    ((NSDescription)creatureArray[i]).runTrial
                        (mapfile, (NSDescription)creatureArray[j]);

                    afterTrial();
                }
            }
        }        
    }

    private void runAllMZTrials(String mapfile) {
        for (int i = 0; i < NUM_MZ_REPEATS; ++i) {
            for (Description d : creatureArray) {
                ((MZDescription)d).runTrial(mapfile);
                afterTrial();
            }
        }
    }

    private synchronized void afterTrial() {
        // Update the data
        Arrays.sort(sortedCreatureArray);
        resultData.fireTableChanged(new TableModelEvent(resultData));
        ++currentTrial;
        progressBar.setValue(currentTrial);
    }

    private void makeGUI() {
        resultData = new ResultData();
        JTable resultTable = new JTable(resultData);

        resultTable.setRowHeight(70);
        resultTable.setShowHorizontalLines(true);
        resultTable.getColumnModel().getColumn(RANK_COL).setPreferredWidth(30);
        resultTable.getColumnModel().getColumn(AUTHOR_COL).setPreferredWidth(120);
        resultTable.getColumnModel().getColumn(IMAGE_COL).setPreferredWidth(45);

        DefaultTableCellRenderer timeRenderer = new DefaultTableCellRenderer();
        timeRenderer.setHorizontalAlignment(SwingConstants.RIGHT);
        timeRenderer.setFont(new Font("Monospaced", Font.PLAIN, 14));

        resultTable.getColumnModel().getColumn(BEST_COL).setCellRenderer(timeRenderer);
        resultTable.getColumnModel().getColumn(WORST_COL).setCellRenderer(timeRenderer);
        if (isMaze) {
            resultTable.getColumnModel().getColumn(MEAN_COL).setCellRenderer(timeRenderer);
        }

        progressBar = new JProgressBar(currentTrial, numTrials);

        resultPicture = new JLabel(this);
        resultArea = new JTextArea(5, 30);

        JPanel resultPane = new JPanel();
        resultScroll = new JScrollPane(resultArea);

        resultPane.setLayout(new BoxLayout(resultPane, BoxLayout.PAGE_AXIS));
        resultPane.add(resultScroll);
        resultPane.add(new JScrollPane(resultPicture));

        JPanel pane = new JPanel();
        pane.setLayout(new BorderLayout());

        JPanel center = new JPanel();
        center.setLayout(new BoxLayout(center, BoxLayout.LINE_AXIS));
        resultTable.setMinimumSize(new Dimension(400, 100));
        center.add(new JScrollPane(resultTable));
        center.add(resultPane);
        pane.add(center, BorderLayout.CENTER);
        pane.add(new JScrollPane(progressBar), BorderLayout.PAGE_END);

        getContentPane().add(pane);
    }
    
    void showText(String s) {
        resultArea.append(s);
        // Move the cursor to the beginning of the last line so that it scrolls nicely
        String contents = resultArea.getText();
        int i = contents.lastIndexOf('\n', contents.length() - 2) + 1;
        resultArea.setCaretPosition(Math.max(0, i));
    }

    // For ResultData
    static final private int RANK_COL     = 0;
    static final private int IMAGE_COL    = 1;
    static final private int CREATURE_COL = 2;
    static final private int AUTHOR_COL   = 3;
    static final private int BEST_COL     = 4;
    static final private int WORST_COL    = 5;
    static final private int MEAN_COL     = 6;

    static final private int WLD_COL      = 4;
    static final private int SCORE_COL    = 5;

    /** These are for maze maps; the natural selection map names are
     * set in the Tournament constructor.*/
    static final private String[] columnNames = 
    {"Rank", "Image", "Creature", "Author", "Best Time", "Worst Time", "Mean Time"};

    private class ResultData extends AbstractTableModel {
        public String getColumnName(int col) {
            return columnNames[col];
        }

        public int getColumnCount() {
            if (isMaze) {
                return columnNames.length; 
            } else {
                return columnNames.length - 1; 
            }
        }

        public Class getColumnClass(int col) {
            if (col == IMAGE_COL) {
                return Icon.class;
            } else {
                return String.class;
            }
        }

        public int getRowCount() { 
            return creatureArray.length;
        }

        public Object getValueAt(int row, int col) {
            Description d = sortedCreatureArray[row];
            switch (col) {
            case RANK_COL:
                return "" + (row + 1);

            case CREATURE_COL:
                return d.className;

            case IMAGE_COL:
                return d.icon;

            case AUTHOR_COL:
                return d.authorName;

            case BEST_COL: // also WLD_COL
                if (isMaze) {
                    if (((MZDescription)d).bestTime == Integer.MAX_VALUE) {
                        return "";
                    } else {
                        return "" + ((MZDescription)d).bestTime;
                    }
                } else {
                    NSDescription n = (NSDescription)d;
                    return "" + n.win + ":" + n.draw + ":" + n.lose;
                }

            case WORST_COL: // also SCORE_COL
                if (isMaze) {
                    if (((MZDescription)d).bestTime == Integer.MAX_VALUE) {
                        return "";
                    } else {
                        return "" + ((MZDescription)d).worstTime;
                    }
                } else {
                    return "" + ((NSDescription)d).score;
                }

            case MEAN_COL:
                if (((MZDescription)d).bestTime == Integer.MAX_VALUE) {
                    return "";
                } else {
                    return "" + ((MZDescription)d).meanTime;
                }
                
            }

            return "";
        }
    }

    public static void main(String[] arg) {
        String mapfile = arg[0];
        String[] creatures = new String[arg.length - 1];
        for (int c = 1; c < arg.length; ++c) {
            creatures[c - 1] = arg[c];
        }

        new Tournament(mapfile, creatures);
    }

    
}
